@@ -9,11 +9,13 @@ from hdbscan import HDBSCAN
 from imblearn.ensemble import BalancedRandomForestClassifier
 from natsort import natsorted
 from scanpy.neighbors import Neighbors
-from sklearn.metrics import confusion_matrix, pairwise_distances, balanced_accuracy_score
+from sklearn.metrics import confusion_matrix, pairwise_distances, balanced_accuracy_score, adjusted_rand_score
 import networkx as nx
 import matplotlib.pyplot as plt
 import seaborn as sns
 from sklearn.model_selection import cross_val_predict
+from functools import lru_cache
+from ..plot import categorical_scatter
 
 
 def _r1_normalize(cmat):
@@ -97,111 +99,55 @@ def _split_train_test_per_group(x, y, frac, max_train, random_state):
     return x_train, y_train, x_test, y_test
 
 
-def _single_supervise_evaluation(
-        clf,
-        x_train,
-        y_train,
-        x_test,
-        y_test,
-        r1_norm_step=0.01,
-        r2_norm_step=0.001,
-        plot=True):
-    # fit model
-    clf.fit(x_train, y_train)
-
-    # calc accuracy
-    y_train_pred = clf.predict(x_train)
-    accuracy_train = balanced_accuracy_score(y_true=y_train, y_pred=y_train_pred)
-    print(f"Balanced accuracy on the training set: {accuracy_train:.3f}")
-    y_test_pred = clf.predict(x_test)
-    accuracy_test = balanced_accuracy_score(y_true=y_test, y_pred=y_test_pred)
-    print(f"Balanced accuracy on the hold-out set: {accuracy_test:.3f}")
-
-    # get confusion matrix
-    y_pred = clf.predict(x_test)
-    cmat = confusion_matrix(y_test, y_pred)
-
-    # normalize confusion matrix
-    r1_cmat = _r1_normalize(cmat)
-    r2_cmat = _r2_normalize(cmat)
-    m1 = np.max(r1_cmat)
-    if np.isnan(m1):
-        m1 = 1.
-    m2 = np.max(r2_cmat)
-    r1_norm_cutoff = m1 - r1_norm_step
-    r2_norm_cutoff = m2 - r2_norm_step
-
-    # final binary matrix to calculate which clusters need to be merged
-    cluster_map = {}
-    judge = np.maximum.reduce([(r1_cmat > r1_norm_cutoff),
-                               (r2_cmat > r2_norm_cutoff)])
-    if judge.sum() > 0:
-        rows, cols = np.where(judge)
-        edges = zip(rows.tolist(), cols.tolist())
-        g = nx.Graph()
-        g.add_edges_from(edges)
-        for comp in nx.connected_components(g):
-            to_label = comp.pop()
-            for remain in comp:
-                cluster_map[remain] = to_label
-
-    if plot:
-        fig, axes = plt.subplots(figsize=(12, 3), dpi=300, ncols=4)
-        ax = axes[0]
-        sns.heatmap(np.log1p(cmat), ax=ax)
-        ax.set_title(f'confusion matrix')
-        ax = axes[1]
-        sns.heatmap(r1_cmat, ax=ax)
-        ax.set_title(f'R1 Norm. cutoff {r1_norm_cutoff:.3f}')
-        ax = axes[2]
-        sns.heatmap(r2_cmat, ax=ax)
-        ax.set_title(f'R2 Norm. cutoff {r2_norm_cutoff:.3f}')
-        ax = axes[3]
-        sns.heatmap(judge, ax=ax)
-        ax.set_title(f'Merge')
-        fig.tight_layout()
-        fig.show()
-    return clf, accuracy_test, cluster_map
-
-
 class ConsensusClustering:
     def __init__(self,
                  model=None,
                  n_neighbors=25,
                  metric='euclidean',
                  min_cluster_size=10,
-                 consensus_rate=0.9,
-                 repeats=200,
-                 resolution=1,
+                 leiden_repeats=200,
+                 leiden_resolution=1,
+                 consensus_rate=0.5,
                  random_state=0,
                  train_frac=0.5,
                  train_max_n=500,
-                 max_iter=10,
-                 target_score=0.95,
-                 n_jobs=-1,
-                 outlier_cell_cutoff=0.25,
-                 outlier_label='Outlier',
-                 plot=True):
+                 max_iter=20,
+                 n_jobs=-1):
+        """
+        Perform consensus clustering by multi-leiden clustering + supervised model
+
+        Parameters
+        ----------
+        model
+        n_neighbors
+        metric
+        min_cluster_size
+        consensus_rate
+        leiden_repeats
+        leiden_resolution
+        random_state
+        train_frac
+        train_max_n
+        max_iter
+        n_jobs
+        """
         # input metrics
         self.min_cluster_size = min_cluster_size
-        self.consensus_rate = consensus_rate
-        self.leiden_repeats = repeats
-        self.leiden_resolution = resolution
+        self.consensus_rate = consensus_rate  # this prevents merging gradient clusters
+        self.leiden_repeats = leiden_repeats
+        self.leiden_resolution = leiden_resolution
         self.random_state = random_state
         self.n_jobs = n_jobs
         self.n_neighbors = n_neighbors
         self.knn_metric = metric
         self.train_frac = train_frac
         self.train_max_n = train_max_n
-        self.plot = plot
         self.max_iter = max_iter
-        self.target_score = target_score
-        self.outlier_cell_cutoff = outlier_cell_cutoff
-        self.outlier_label = outlier_label
-
+        self.minimum_n_cluster = None  # estimated from multi leiden mean
         self.n_obs, self.n_pcs = None, None
         self.X = None
         self._neighbors = None
+        self.step_data = {}
 
         # multiple leiden clustering
         self.leiden_result_df = None
@@ -212,6 +158,7 @@ class ConsensusClustering:
         self._label_with_leiden_outliers = None
         self.label = None
         self.label_proba = None
+        self.cv_predicted_label = None
         self.final_accuracy = None
         return
 
@@ -221,7 +168,7 @@ class ConsensusClustering:
 
         # Construct KNN graph
         print('Computing nearest neighbor graph')
-        self.compute_neighbors()
+        self._compute_neighbors()
 
         # repeat Leiden clustering with different random seeds
         print('Computing multiple clustering with different random seeds')
@@ -230,15 +177,15 @@ class ConsensusClustering:
             leiden_kwds = kwds
         else:
             leiden_kwds.update(kwds)
-        self.multi_leiden_clustering(**leiden_kwds)
+        self._multi_leiden_clustering(**leiden_kwds)
 
         # merge the over clustering version by supervised learning
-        self.supervise_learning()
+        self._supervise_learning()
 
         # assign outliers
-        self.final_evaluation()
+        self._final_evaluation()
 
-    def compute_neighbors(self):
+    def _compute_neighbors(self):
         # nearest neighbors graph
         adata = anndata.AnnData(X=None,
                                 obs=pd.DataFrame([], index=[f'obs{i}' for i in range(self.n_obs)]),
@@ -256,11 +203,11 @@ class ConsensusClustering:
             random_state=self.random_state)
         return
 
-    def multi_leiden_clustering(self,
-                                partition_type=None,
-                                partition_kwargs=None,
-                                use_weights=True,
-                                n_iterations=-1):
+    def _multi_leiden_clustering(self,
+                                 partition_type=None,
+                                 partition_kwargs=None,
+                                 use_weights=True,
+                                 n_iterations=-1):
         """Modified from scanpy"""
         if self._neighbors is None:
             raise ValueError('Run compute_neighbors first before multi_leiden_clustering')
@@ -319,13 +266,13 @@ class ConsensusClustering:
         cluster_count = self.leiden_result_df.apply(lambda i: i.unique().size)
         print(f'Found {cluster_count.min()} - {cluster_count.max()} clusters, '
               f'mean {cluster_count.mean():.1f}, std {cluster_count.std():.2f}')
-
+        self.minimum_n_cluster = max(1, int(cluster_count.mean() * 0.8))
         # create a over-clustering version based on all the leiden runs
         print('Summarizing multiple clustering results')
-        self.summarize_multi_leiden()
+        self._summarize_multi_leiden()
         return
 
-    def summarize_multi_leiden(self):
+    def _summarize_multi_leiden(self):
         # here we don't rely on hdbscan clustering, just use it to reduce the pairwise distances calculation
         # the resulting clusters are simply disconnected components based on hamming dist graph
         # This results an over-clustering, which will be merged by the supervise learning step.
@@ -357,8 +304,8 @@ class ConsensusClustering:
                         clusters[node] = -1
 
         clusters = pd.Series(clusters).sort_index()
-        print(f'{(clusters != -1).sum()} cells assigned to {clusters.unique().size} raw clusters')
-        print(f'{(clusters == -1).sum()} cells are outliers')
+        print(f'{(clusters != -1).sum()} cells assigned to {clusters.unique().size - 1} raw clusters')
+        print(f'{(clusters == -1).sum()} cells are multi-leiden outliers')
         self._multi_leiden_clusters = clusters.values
         return
 
@@ -385,7 +332,72 @@ class ConsensusClustering:
             class_weight=None)
         return clf
 
-    def supervise_learning(self):
+    @lru_cache(999)
+    def _cluster_pairs_hamming_dist(self, labels_tuple, c1, c2):
+        labels = np.array(labels_tuple)
+        return pairwise_distances(self.leiden_result_df.loc[labels == c1],
+                                  self.leiden_result_df.loc[labels == c2],
+                                  metric='hamming').mean()
+
+    def _single_supervise_evaluation(
+            self,
+            clf,
+            cur_y,
+            x_train,
+            y_train,
+            x_test,
+            y_test,
+            r1_norm_step=0.01,
+            r2_norm_step=0.001):
+        # fit model
+        clf.fit(x_train, y_train)
+
+        # calc accuracy
+        y_train_pred = clf.predict(x_train)
+        accuracy_train = balanced_accuracy_score(y_true=y_train, y_pred=y_train_pred)
+        print(f"Balanced accuracy on the training set: {accuracy_train:.3f}")
+        y_test_pred = clf.predict(x_test)
+        accuracy_test = balanced_accuracy_score(y_true=y_test, y_pred=y_test_pred)
+        print(f"Balanced accuracy on the hold-out set: {accuracy_test:.3f}")
+
+        # get confusion matrix
+        y_pred = clf.predict(x_test)
+        cmat = confusion_matrix(y_test, y_pred)
+
+        # normalize confusion matrix
+        r1_cmat = _r1_normalize(cmat)
+        r2_cmat = _r2_normalize(cmat)
+        m1 = np.max(r1_cmat)
+        if np.isnan(m1):
+            m1 = 1.
+        m2 = np.max(r2_cmat)
+
+        cluster_map = {}
+        init_m1 = m1
+        init_m2 = m2
+        while (len(cluster_map) == 0) and (m1 > init_m1 * 0.5) and (m2 > init_m2 * 0.5):
+            m1 -= r1_norm_step
+            m2 -= r2_norm_step
+
+            # final binary matrix to calculate which clusters need to be merged
+            judge = np.maximum.reduce([(r1_cmat > m1),
+                                       (r2_cmat > m2)])
+            if judge.sum() > 0:
+                rows, cols = np.where(judge)
+                edges = zip(rows.tolist(), cols.tolist())
+                g = nx.Graph()
+                g.add_edges_from(edges)
+                for comp in nx.connected_components(g):
+                    to_label = comp.pop()
+                    for remain in comp:
+                        cluster_dist = self._cluster_pairs_hamming_dist(tuple(cur_y), remain, to_label)
+                        if cluster_dist < self.consensus_rate:
+                            print(f'Distance between {remain} - {to_label}: '
+                                  f'{cluster_dist:.3f} < {self.consensus_rate:.3f}')
+                            cluster_map[remain] = to_label
+        return clf, accuracy_test, cluster_map, cmat, r1_cmat, r2_cmat
+
+    def _supervise_learning(self):
         if self._multi_leiden_clusters is None:
             raise ValueError('Run multi_leiden_clustering first to get a '
                              'clustering assignment before run supervise_learning.')
@@ -419,16 +431,20 @@ class ConsensusClustering:
                 max_train=self.train_max_n,
                 random_state=self.random_state + cur_iter  # every time train-test split got a different random state
             )
-            clf, score, cluster_map = _single_supervise_evaluation(clf,
-                                                                   x_train,
-                                                                   y_train,
-                                                                   x_test,
-                                                                   y_test,
-                                                                   r1_norm_step=0.01,
-                                                                   r2_norm_step=0.001)
-            if self.target_score < score:
-                print('Stop iteration because score > target score.')
-                break
+            clf, score, cluster_map, cmat, r1_cmat, r2_cmat = self._single_supervise_evaluation(
+                clf,
+                cur_y,
+                x_train,
+                y_train,
+                x_test,
+                y_test,
+                r1_norm_step=0.01,
+                r2_norm_step=0.001)
+
+            # save step data for plotting
+            self.step_data[cur_iter] = [cur_y, cmat, r1_cmat, r2_cmat, cluster_map]
+
+            # judge results
             if len(cluster_map) > 0:
                 print(f'Merging {len(cluster_map)} clusters.')
                 cur_y = pd.Series(cur_y).apply(lambda i: cluster_map[i] if i in cluster_map else i)
@@ -438,6 +454,10 @@ class ConsensusClustering:
             else:
                 print('Stop iteration because there is no cluster to merge')
                 break
+            if len(set(cur_y.tolist())) <= self.minimum_n_cluster:
+                print('Stop iteration because # of clusters is much smaller '
+                      'than average # of clusters in leiden clustering.')
+                break
         else:
             print('Stop iteration because reaching maximum iteration.')
         self._label_with_leiden_outliers = cur_y
@@ -446,23 +466,8 @@ class ConsensusClustering:
         self.final_accuracy = score
         return
 
-    def final_evaluation(self):
+    def _final_evaluation(self):
         print(f'\n=== Assign final labels ===')
-        # final evaluation of non-outliers using cross val predict
-        _x = self.X[self.label != -1]
-        _idx = np.where(self.label != -1)[0]
-        _y = self.label[self.label != -1]
-        final_predict_proba = cross_val_predict(self.supervise_model,
-                                                _x,
-                                                y=_y,
-                                                method='predict_proba',
-                                                n_jobs=self.n_jobs,
-                                                verbose=0,
-                                                cv=10)
-        final_predict = pd.Series(np.argmax(final_predict_proba, axis=1), index=_idx)
-        final_cell_proba = pd.Series(np.max(final_predict_proba, axis=1), index=_idx)
-        final_acc = balanced_accuracy_score(_y, final_predict.values)
-        print(f'Final CV Accuracy on non-outliers: {final_acc:.3f}')
 
         # predict outliers
         outlier_x = self.X[self.label == -1]
@@ -470,21 +475,123 @@ class ConsensusClustering:
         if len(outlier_idx) != 0:
             outlier_predict = pd.Series(self.supervise_model.predict(outlier_x),
                                         index=outlier_idx)
-            outlier_proba = self.supervise_model.predict_proba(outlier_x)
-            outlier_cell_proba = pd.Series(outlier_proba.max(axis=1),
-                                           index=outlier_idx)
-            final_predict = pd.concat([final_predict, outlier_predict]).sort_index()
-            final_cell_proba = pd.concat([final_cell_proba, outlier_cell_proba]).sort_index()
+            for cell, pred_label in outlier_predict.items():
+                self.label[cell] = pred_label
+        print(f'Assigned all the multi-leiden clustering outliers into clusters '
+              f'using the prediction model from final clustering version.')
 
-        # final outlier assignment by cell-level max proba < outlier_cell_cutoff
-        final_predict[final_cell_proba.values < self.outlier_cell_cutoff] = -1
-        print(f'Assign {(final_predict == -1).sum()} cells as outliers '
-              f'based on cell-level max prediction probability < {self.outlier_cell_cutoff:.3f}')
+        # final evaluation of non-outliers using cross val predict
+        final_predict_proba = cross_val_predict(self.supervise_model,
+                                                self.X,
+                                                y=self.label,
+                                                method='predict_proba',
+                                                n_jobs=self.n_jobs,
+                                                verbose=0,
+                                                cv=10)
+        final_predict = pd.Series(np.argmax(final_predict_proba, axis=1))
+        final_cell_proba = pd.Series(np.max(final_predict_proba, axis=1))
+        final_acc = balanced_accuracy_score(self.label, final_predict.values)
+        print(f'Final ten-fold CV Accuracy on all the cells: {final_acc:.3f}')
 
-        self.label = final_predict.apply(lambda i: f'c{i:03d}' if i != -1 else self.outlier_label).values
+        self.label = [f'c{label}' for label in self.label]
+        self.cv_predicted_label = [f'c{label}' for label in final_predict]
         self.label_proba = final_cell_proba.values
         self.final_accuracy = final_acc
         return
 
     def save(self, output_path):
         joblib.dump(self, output_path)
+
+    def plot_leiden_cases(self,
+                          coord_data,
+                          coord_base='umap',
+                          plot_size=3,
+                          dpi=300,
+                          plot_n_cases=4,
+                          s=3):
+        # choose some most different leiden runs by rand index
+        sample_cells = min(1000, self.leiden_result_df.shape[0])
+        sample_runs = min(30, self.leiden_result_df.shape[1])
+        use_df = self.leiden_result_df.sample(sample_cells).T.sample(sample_runs)
+        rand_index_rank = pd.DataFrame(
+            pairwise_distances(use_df, metric=adjusted_rand_score),
+            index=use_df.index,
+            columns=use_df.index).unstack().sort_values()
+        plot_cases = set()
+        for pairs in rand_index_rank[:10].index:
+            plot_cases.add(pairs[0])
+            plot_cases.add(pairs[1])
+            if len(plot_cases) > plot_n_cases:
+                break
+        plot_cases = list(plot_cases)[:plot_n_cases]
+
+        # plot
+        plot_data = coord_data.copy()
+        fig, axes = plt.subplots(figsize=(plot_n_cases * plot_size, plot_size),
+                                 ncols=plot_n_cases,
+                                 dpi=dpi)
+
+        for case, ax in zip(plot_cases, axes):
+            plot_data[f'Leiden {case}'] = self.leiden_result_df[case].values
+            categorical_scatter(ax=ax,
+                                data=plot_data,
+                                coord_base=coord_base,
+                                hue=f'Leiden {case}',
+                                s=s)
+        return fig, axes
+
+    def plot_steps(self, coord_data, coord_base='umap', plot_size=3, dpi=300):
+        plot_data = coord_data.copy()
+
+        # initial clusters
+        for i, step in enumerate(sorted(self.step_data.keys())):
+            cur_y, cmat, r1_cmat, r2_cmat, cluster_map = self.step_data[step]
+            if i == 0:
+                fig, ax = plt.subplots(figsize=(3, 3), dpi=dpi)
+                plot_data['cur_y'] = cur_y
+                _ = categorical_scatter(data=plot_data,
+                                        ax=ax,
+                                        hue='cur_y',
+                                        coord_base=coord_base,
+                                        palette='tab20',
+                                        text_anno='cur_y',
+                                        show_legend=False)
+                ax.set(title='Initial Labels From\nMulti-leiden Clustering')
+
+            fig, axes = plt.subplots(figsize=(4 * plot_size, plot_size),
+                                     ncols=4,
+                                     dpi=dpi)
+            ax = axes[0]
+            sns.heatmap(ax=ax, data=cmat, cbar=None, cmap='Reds')
+            ax.set(title='Confusion Matrix', ylabel=f'Step {step}')
+            ax = axes[1]
+            sns.heatmap(ax=ax, data=r1_cmat, cbar=None, cmap='Reds')
+            ax.set(title='R1 Norm.')
+            ax = axes[2]
+            sns.heatmap(ax=ax, data=r2_cmat, cbar=None, cmap='Reds')
+            ax.set(title='R2 Norm.')
+            ax = axes[3]
+            if len(cluster_map) > 0:
+                involved_clusters = set(
+                    list(cluster_map.keys()) + list(cluster_map.values()))
+                cur_y = pd.Series(cur_y, index=plot_data.index)
+                cur_y = cur_y.apply(lambda i: cluster_map[i] if i in cluster_map else i)
+                # if not involved, mark as -1
+                cur_y = cur_y.apply(lambda i: i if i in involved_clusters else -1)
+                plot_data['cur_y'] = cur_y
+                n_color = cur_y.unique().size - 1
+                colors = list(sns.color_palette('tab10', n_color))
+                cmap = {
+                    c: colors.pop() if c != -1 else (0.9, 0.9, 0.9)
+                    for c in cur_y.unique()
+                }
+                categorical_scatter(ax=ax,
+                                    data=plot_data,
+                                    coord_base=coord_base,
+                                    hue='cur_y',
+                                    palette=cmap,
+                                    s=3)
+                ax.set(title=f'Step {step} Cells After Merge')
+            else:
+                ax.axis('off')
+        return
